﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Text;
using Fadd.IO.Logging.Targets.File;

namespace Fadd.Logging.Targets
{
    public class FileTarget : ILogTarget
    {
        IFileWriter _fileWriter;

        public FileTarget(string name, FileConfiguration configuration)
        {
            _fileWriter = new FileWriter(name, configuration);
        }

        public FileTarget(IFileWriter writer)
        {
            _fileWriter = writer;
        }

        protected FileConfiguration Configuration
        {
            get { return _fileWriter.Configuration; }
        }


        protected virtual string FormatException(Exception exception, int itendentation)
        {
            string text = "\r\n******* EXCEPTION ********"
                          + exception.ToString().Replace("\r\n", "\r\n".PadRight(itendentation, '\t'));
            if (exception.InnerException != null)
                return text + FormatException(exception.InnerException, itendentation + 1);
            return text;
        }

        protected virtual string FormatLogEntry(LogEntry entry)
        {
            if (entry.Exception != null)
            {
                return string.Format("{0}|{1}|{2}|{3}|{4}|{5}\r\n{6}\r\n",
                                     entry.CreatedAt.ToString(Configuration.DateTimeFormat),
                                     entry.LogLevel,
                                     entry.ThreadId,
                                     FormatUserName(entry.UserName, 40),
                                     FormatStackTrace(entry.StackFrames, 100),
                                     FormatMessage(entry.Message),
                                     FormatException(entry.Exception, 1)
                    );
            }

            return string.Format("{0}|{1}|{2}|{3}|{4}|{5}\r\n",
                                 entry.CreatedAt.ToString(Configuration.DateTimeFormat),
                                 entry.LogLevel,
                                 entry.ThreadId,
                                 FormatUserName(entry.UserName, 40),
                                 FormatStackTrace(entry.StackFrames, 100),
                                 FormatMessage(entry.Message)
                );
        }

        protected virtual string FormatMessage(string message)
        {
            return message.Replace("\r\n", "\r\n\t").Replace('|', ';');
        }

        protected virtual string FormatStackTrace(StackFrame[] frames, int maxSize)
        {
            var typeName = frames[0].GetMethod().ReflectedType.Name;
            var methodName = frames[0].GetMethod().Name;
            var result = string.Format("{0}.{1}", typeName, methodName);
            if (result.Length < maxSize)
                return result;

            result = methodName;
            if (result.Length < maxSize)
                return result;

            return result.Substring(0, maxSize - 1) + ".";
        }

        protected virtual string FormatUserName(string userName, int maxSize)
        {
            if (userName.Length > 0)
            {
                int pos = userName.IndexOf('\\'); //domain name
                if (pos != -1)
                    userName = userName.Substring(pos + 1);
            }

            return userName.Length > maxSize ? userName.Substring(0, maxSize) : userName;
        }

        protected string MaxSize(string tmp, int size)
        {
            if (tmp.Length > size)
                return tmp.Substring(0, size - 1) + ".";
            return tmp;
        }


        public string Name
        {
            get { return _fileWriter.Name; }
        }

        public void Enqueue(LogEntry entry)
        {
            var entryString = FormatLogEntry(entry);
            _fileWriter.Write(entryString);
        }
    }
}